/**
 * @author Peter Fricke
 */
package hitters.multi;

import hitters.tools.LogService;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

/**
 * @author pp
 *
 */
public abstract class AbstractHHH{

	Parameter myParams;
	int N          = 0;
	int tupelCount = 0;
	int maxTupel   = 0;	


	public AbstractHHH( Parameter myParams ){

		this.myParams = myParams;

		if( myParams.getDim() != 2 ) throw new RuntimeException("Nur 2" +
		" dim durchgängig implementiert.");

		if( LogService.shouldLog( LogService.NORMAL ) ) 
			LogService.log( LogService.NORMAL, 
					"AbstractHHH: DimI = " + myParams.getDimI() +
					" Dims = " + myParams.getDimS() + 
					" Dim = " + myParams.getDim() + 
					" und myParams.getL() +1 " +  
					(myParams.getL() +1) );		
	}


	// Berechnet die Menge der undominierten HHH unterhalb eines Labels.
	// Erlaubt schnelleren Output durch Reduzieren der
	// Menge der HHH, die fuer das Diskontieren überhaupt 
	// geprüft werden müssen.
	public Set<Element> trimHitters( 
			Set<Element> fullHitters, int[] label ){
		
		boolean blocked = false;
		HashSet<Element> temp =	new HashSet<Element>();
		HashSet<Element> trim =	new HashSet<Element>();
		
		// Nur Nachfahren von Elementen des Labels
		for( Element h : fullHitters ){
			if( h.isBelowLabel( label ) ) temp.add(h);
		}
		
		// Nur undominierte HHH
		for( Element h : temp ){
			blocked = false;
			for( Element other : temp ){
				if( other.equals(h) == false && other.isGenOf(h) == true )
					blocked = true;
			}
			if( ! blocked ) trim.add(h);
		}
		
		return trim;
	}


	public void insert(Element e, int count){
		insert( e, count, true );
	}
	
	public abstract void insert(Element e, int count, boolean increase);

	public abstract HashMap<Element, MultiHitterInfo> outputSet(double phi);	

	// Zum Loggen der absoluten und diskontierten Haeufigkeiten
	public abstract HashMap<Element, MultiHitterInfo> outputSet(double phi, boolean log);

	public abstract HashMap<Element, Integer> dumpf();

	public abstract HashMap<Element, Integer> dumpF(double phi);

	public abstract int instantiateDelta(Element e);

	public int getN() { return N; }

	public int getTupelCount() { return tupelCount; }

	public int getMaxTupelCount() { return Math.max( maxTupel, tupelCount); }

	public Parameter getParameter() { return myParams; }

}
