/**
 * 
 */
package hitters.multi;

import hitters.tools.LogService;

import java.io.*;
import java.util.*;


/**
 * @author Peter Fricke
 * 
 */
public class MultiDatabase {

	final int HITTER = 1;
	final int NODES  = 2;
	private String filename;
	private FileWriter outData;
	private BufferedWriter out;
	private FileReader inData;
	private BufferedReader in;
	private int N   = -1;
	private double phi     = -1;
	private double epsilon = -1;
	private Parameter myParams; 

	public MultiDatabase( String filename, Parameter myParams ){

		if( LogService.shouldLog( LogService.NORMAL ) ) 
			LogService.log(LogService.NORMAL,  "MultiDataBase:" +
					" dimI lt. Par " + myParams.getDimI() +
					" dimS lt Par " + myParams.getDimS() );

		this.filename  = filename;
		this.myParams  = myParams;	
	}


	public Vector<TestCaseInfo> readTestCaseInfo(){		

		int dim = -1;
		int dimS = -1, dimI = -1; //Tatsächliche Dim in Datei (vs. myParams)
		int number; // Anzahl der Elemente mit dem angegebenen Präfix 
		String line;
		String[] item;		
		String[] clean;		
		String[] eStr;		
		String[] eInt;		

		Element e;
		Vector<TestCaseInfo> data = new Vector<TestCaseInfo>();

		try{ 			
			while( (line = in.readLine()) != null ){
				if( line.length() > 0 ){		
					if( line.startsWith("#$") ){

						line = line.substring( 2 );
						dimS = Integer.parseInt( line.split("[|]")[0].trim());
						dimI = Integer.parseInt( line.split("[|]")[1].trim());						

						if( dimS != myParams.getDimS() || dimI != myParams.getDimI() ){

							if( LogService.shouldLog( LogService.NORMAL ) )
								LogService.log( LogService.NORMAL, 
										"Dimensionen in Datei entsprechen" +
								" nicht Dimensionen im Parameterobjekt.");

							dimS = myParams.getDimS();
							dimI = myParams.getDimI();							
						}

						dim = dimI + dimS;
					}
					if( line.startsWith("#@") ){						

						if( dim < 1 ) throw new RuntimeException(" Dimensionen" +
						" nicht in Testdatei angegeben oder null.");

						line = line.substring( 2 );						
						item = line.split( "[|]" );

						if( dim != item.length - 1 ) 
							throw new RuntimeException( "Verschiedene Dim." +
							" beim Lesen der Daten." ); 

						clean = new String[ item.length - 1 ];

						for( int i = 0; i < item.length - 1; i++ ) 
							clean[i] = item[i].trim();

						number = Integer.parseInt( item[ item.length - 1].trim() );

						eStr = new String[ dimS ];  
						eInt = new String[ dimI ];

						for( int i = 0; i < dimS; i++ ) eStr[i] = clean[i];
						for( int i = 0; i < dimI; i++ ){
							eInt[i] = clean[ i + dimS ]; 
						}
						e = Element.createElement( eStr, eInt, myParams );
						data.add( new TestCaseInfo( e, number ) );
					}
				}
			}	    
		}		
		catch (java.io.IOException ex) { System.out.println(ex); };	   

		return data;
	}


	public Vector<Element> readElements(){ 

		int dim = -1;
		int dimS = -1, dimI = -1; //Tatsächliche Dim in Datei (vs. myParams) 
		String line;
		String[] item;		
		String[] clean;		
		String[] eStr;		
		String[] eInt;				
		Element e;
		Vector<Element> data = new Vector<Element>();    	    

		try{ 
			while( (line = in.readLine()) != null ){      
				if( line.length() > 0 ){		
					if( line.startsWith("#$") ){

						line = line.substring( 2 );
						dimS = Integer.parseInt( line.split("[|]")[0].trim());
						dimI = Integer.parseInt( line.split("[|]")[1].trim());						

						if( dimS != myParams.getDimS() || dimI != myParams.getDimI() ){

							if( LogService.shouldLog( LogService.NORMAL ) )
								LogService.log( LogService.NORMAL, 
										"Dimensionen in Datei entsprechen " +
								"nicht Dimensionen im Parameterobjekt.");

							dimS = myParams.getDimS();
							dimI = myParams.getDimI();							
						}

						dim = dimI + dimS;
					}
					if( line.startsWith("/") ){						

						if( dim < 1 ) throw new RuntimeException( "Dimensionen " +
						"nicht in Datei angegeben oder null.");

						item = line.split( "[|]" );
						if( dim != item.length  ) 
							throw new RuntimeException( "Verschiedene Dim." +
							" beim Lesen der Daten." ); 

						clean = new String[ item.length ];
						for( int i = 0; i < item.length; i++ ) clean[i] = item[i].trim();
						eStr = new String[ dimS ];  
						eInt = new String[ dimI ];

						for( int i = 0; i < dimS; i++ ) eStr[i] = clean[i];
						for( int i = 0; i < dimI; i++ ) eInt[i] = clean[ i + dimS ]; 

						e = Element.createElement( eStr, eInt, myParams );
						data.add( e.clone() );
					}
				}
			}
		}		
		catch (java.io.IOException ex) {};	   

		return data;
	}


	public Vector<Element> readSystemCalls(){ 

		/*
		int dim = -1;
		int dimS = -1, dimI = -1;  		
		String[] item;		
		String[] clean;		
		String[] eStr;		
		String[] eInt;*/				
		String line;
		SysElement e = null;
		Vector<Element> data = new Vector<Element>();    	    
		HashMap<Integer, String> fdTable = new HashMap<Integer, String>(60);

		try{ 
			while( (line = in.readLine()) != null ){      
				if( line.length() > 0 ){		
					/*if( line.startsWith("#$") ){
						line = line.substring( 2 );
						dimS = Integer.parseInt( line.split("[|]")[0].trim());
						dimI = Integer.parseInt( line.split("[|]")[1].trim());
						dim = dimI + dimS;
						//HHHMulti.log( "DimS " + dimS + " dimI " + dimI);						
					}*/
					if( LogService.shouldLog( LogService.NORMAL ) ) 
						LogService.log(LogService.NORMAL, "DATABASE" +
								" lese : System Call " + line ); 

					if( line.startsWith("open(") )  
						e = SysParser.createSystemCallElement( line, fdTable, myParams );
					if( line.startsWith("close(") ) 
						e = SysParser.createSystemCallElement( line, fdTable, myParams );
					if( line.startsWith("read(") )  
						e = SysParser.createSystemCallElement( line, fdTable, myParams );
					if( line.startsWith("write(") ) 
						e = SysParser.createSystemCallElement( line, fdTable, myParams );
					if( line.startsWith("gettimeofday(") ) 
						e = SysParser.createSystemCallElement( line, fdTable, myParams );

					if( e != null ) data.add( e.clone() );
					//					if( e != null ) System.out.println( "DATABASE gelesen : System Call " +
					//							e.simpleToString() + e);
					e = null;
				}
			}
		}		
		catch (java.io.IOException ex) {};	   

		return data;
	}


	public boolean openRead(){
		boolean success = true;
		try{ 
			inData = new FileReader(filename);
			in = new BufferedReader(inData);		
		}
		catch (java.io.FileNotFoundException e){ 
			System.out.print(e.getMessage());
			success = false;
		}	    	    
		return success;
	}


	public void closeRead(){	    
		try{
			in.close();
			inData.close();
		}
		catch (java.io.IOException e) {  };	    
	}


	public boolean openWrite(){
		boolean success = true;
		try{ 
			outData = new FileWriter(filename);
			out = new BufferedWriter(outData);
		}
		catch (java.io.FileNotFoundException e){ 
			System.out.print(e.getMessage());
			success = false;
		}
		catch (java.io.IOException e) { success = false; };

		return success;
	}


	public void closeWrite(){	    
		try{
			out.close();
		}
		catch (java.io.IOException e) { 
			System.out.println("CloseWrite: Error: " + e.getMessage()); 
		};	    
	}


	public void println(String s) {
		try{ 
			out.write( s + System.getProperty("line.separator") );
		}
		catch (java.io.FileNotFoundException e){ 
			System.out.print(e.getMessage());		    	
		}
		catch (java.io.IOException e) { };
	}


	public void reset() {
		closeRead();
		openRead();		
	}


	public double getPhi() {
		if( in != null) reset();
		else openRead();
		if( phi < 0 ) readSolution();
		if( phi < -1E-10 ) throw new RuntimeException("Kann phi nicht lesen");
		closeRead();
		return phi;
	}


	public double getEpsilon() {
		if( in != null) reset();
		else openRead();
		if( epsilon < 0 ) readSolution();

		if( epsilon < -1E-10 ) 
			throw new RuntimeException("Kann epsilon nicht lesen");

		closeRead();
		return epsilon;
	}


	public int getSize() {
		if( in != null ) reset();
		else openRead();
		if( N < 0 ) readSolution();
		if( N < 0 ) throw new RuntimeException("Kann N nicht lesen");
		closeRead();
		return N;
	}


	public HashMap<Element, MultiStringElement> readNodes() {
		HashMap<Element, MultiStringElement> map;
		map = (HashMap<Element, MultiStringElement>) genericReadSolution( NODES );
		return map;
	}


	public HashMap<Element, MultiHitterInfo> readSolution() {
		HashMap<Element, MultiHitterInfo> map;
		map = (HashMap<Element, MultiHitterInfo>) genericReadSolution( HITTER );
		return map;
	}


	private HashMap genericReadSolution( int type ) {

		HashMap solution = null;
		String marker    = null;

		switch( type ){

		case HITTER:
			marker = "#=";
			solution = new HashMap<Element, MultiHitterInfo>();
			break;

		case NODES:
			marker = "#*";
			solution = new HashMap<Element, MultiStringElement>();
			break;

		}

		int dim = -1;
		int dimS = -1, dimI = -1; //Tatsächliche Dim in Datei (vs. myParams)
		//int number; // Anzahl der Elemente mit dem angegebenen Präfix 
		String line;
		String[] item;		
		String[] clean;		
		String[] eStr;		
		String[] eInt;		

		Element e;
		int i1, i2, i3;

		try{ 			
			while( (line = in.readLine()) != null ){
				if( line.length() > 0 ){	
					if( line.startsWith("#$") ){

						line = line.substring( 2 );
						dimS = Integer.parseInt( line.split("[|]")[0].trim());
						dimI = Integer.parseInt( line.split("[|]")[1].trim());						

						if( dimS != myParams.getDimS() || dimI != myParams.getDimI() ){
							if( LogService.shouldLog( LogService.NORMAL ) ) 
								LogService.log( LogService.NORMAL, 
										"Dimensionen in Datei entsprechen " +
								"nicht Dimensionen im Parameterobjekt.");

							dimS = myParams.getDimS();
							dimI = myParams.getDimI();							
						}
						dim = dimI + dimS;
					}
					if( line.startsWith("#%") ){

						line 	= line.substring( 2 );
						N 		= Integer.parseInt( line.split("[|]")[0].trim());
						phi 	= Double.parseDouble( line.split("[|]")[1].trim());						
						epsilon = Double.parseDouble( line.split("[|]")[2].trim());
					}
					if( line.startsWith( marker ) ){						

						if( dim < 1 ) throw new RuntimeException("Dimensionen " +
						"nicht in Musterlösung angegeben oder null.");
						if( phi < -1E10 ) throw new RuntimeException("Phi nicht " +
						"in Musterlösung angegeben.");
						if( epsilon < -1E10 ) throw new RuntimeException(" Epsilon " +
						"nicht in Musterlösung angegeben.");

						line = line.substring( 2 );						
						item = line.split( "[|]" );

						if( dim != item.length - 3 ) 
							throw new RuntimeException( "Verschiedene Dim. beim Lesen der Daten." );

						clean = new String[ item.length - 3 ];

						for( int i = 0; i < item.length - 3; i++ ) clean[i] = item[i].trim();
						i1 = Integer.parseInt( item[ item.length - 3 ].trim() );
						i2 = Integer.parseInt( item[ item.length - 2 ].trim() );
						i3 = Integer.parseInt( item[ item.length - 1 ].trim() );

						eStr = new String[ dimS ];  
						eInt = new String[ dimI ];

						for( int i = 0; i < dimS; i++ ) eStr[i] = clean[i];
						for( int i = 0; i < dimI; i++ ){
							eInt[i] = clean[ i + dimS ]; 
						}
						e = Element.createElement( eStr, eInt, myParams );					

						switch( type ){

						case HITTER:							
							solution.put( e, new MultiHitterInfo( i1, i2, i3 ) );
							break;

						case NODES:
							solution.put( e, new MultiStringElement( i1, i2, i3 ) );
							break;

						}						
					}
				}
			}	    
		}		
		catch (java.io.IOException ex) { System.out.println(ex); };	   

		return solution;
	}

}
