package hitters.multi;

import java.util.HashMap;
import java.util.regex.Pattern;

/**
 * @author Peter Fricke
 * 
 */
public class SysParser {

	public static final int MAXDEPTH = 20;

	private static Pattern csv = Pattern.compile(",");


	public static String createString( String[] eStr, 
			int[] eInt, Parameter myParams ){		

		int res;
		int[] code;
		String s = "";
		for( int i = 0; i < eStr.length; i++ ){
			s = s + eStr[i];
			if( i < myParams.getDim() -1 ) s = s + " (s) | ";
			else s = s + " (s)";
		}		
		for( int i = 0; i < eInt.length; i++ ){
			code = new int[MAXDEPTH];			

			for( int j = 1; j < myParams.getMaskLength(i); j++ ){

				res = eInt[i] & 
				myParams.getMask(i,j) & ~ myParams.getMask(i, j-1); 

				res = res >>> ( 32 - myParams.getPlength(i,j) );				
		code[j] = res;				
			}
			s = s + myParams.getTaxonomy().getPathlabel( code );						

			if( i < eInt.length -1 ) s = s + " (i) | ";
			else s = s + " (i)";
		}
		return s;		
	}


	public static SysElement createSystemCallElement(String line, HashMap<Integer, 
			String> fdTable, Parameter myParams ) {

		// Wenn Fehler auftreten, ueberspringe statt Abbbruch,
		// Ursache loggen. Nur im Produktionsbetrieb einschalten!

		//try{			

		//Klammern nicht entfernen: write vs. writev usw.
		if( line.startsWith("open(") )  
			return getOpen(  line, myParams, fdTable );
		if( line.startsWith("close(") ) 
			return getClose( line, myParams, fdTable );
		if( line.startsWith("read(") )  
			return getRead(  line, myParams, fdTable );
		if( line.startsWith("write(") ) 
			return getWrite( line, myParams, fdTable );

		// Standardbehandlung mit Pfad
		if( line.startsWith("fstat64("))
			return getSimpleWPath( line, "/2/4/0/0", myParams, fdTable );

		// Standardbehandlung ohne Pfad
		if( line.startsWith("gettimeofday(") ) 	
			return getSimple( "/2/6/0/0", myParams );
		if( line.startsWith("readlink("))		
			return getSimple( "/2/10/0/0", myParams );
		if( line.startsWith("time("))			
			return getSimple( "/2/13/0/0", myParams );
		if( line.startsWith("uname("))			
			return getSimple( "/2/14/0/0", myParams );

		//		}
		//		catch( Exception e ){
		//			LogService.logF( "# Problem beim Parsen der Systemcalls: " +
		//					"line = " + line + " fuehrt zu Fehler:  " +
		//					e.toString() + " " );			
		//		}

		return null;
	}



	private static SysElement getClose( String s, 
			Parameter myParams, HashMap<Integer, String> fdTable ) {

		SysElement e = null;
		String file  = null;
		//int ret = getReturnValue(s);
		int pos = s.indexOf("(");		
		int end = s.indexOf(")", pos);
		int fd  = Integer.parseInt( s.substring(pos+1, end).trim() );		
		if( fd > -1 ) file = fdTable.get(fd);

		if( file == null ) file = "/DateiUnbekannt";

		fdTable.remove(fd);

		String[] eStr = new String[ myParams.getDimS() ];
		eStr[0] = file;

		String[] eInt = new String[ myParams.getDimI() ];
		for(int i = 1; i < myParams.getDimI(); i++ ) eInt[i] = "";

		eInt[0] = "/1/4/0/0";
		int[] res = Element.stringToInt( eInt, myParams );
		e = new SysElement( eStr, res, myParams );

		return e;
	}


	private static SysElement getOpen( String s, Parameter myParams,
			HashMap<Integer, String> fdTable ) {

		SysElement e = null;		
		int ret = getReturnValue(s);

		// Pfad extrahieren 
		int pos      = s.lastIndexOf("\"");
		String file  = s.substring(6, pos);		
		if( file == null ) file = "/DateiUnbekannt";

		// Tiefe des Pfads begrenzen
		// Pfade vom Typ "/etc//////////sound" normieren
		int lev = 0;
		int k = file.indexOf( '/', 0 );
		int oldk = k;
		while( k >= 0){
			k = file.indexOf( '/', k + 1 );
			if( k == oldk + 1 ){ // Ueberschuessige slashs raus
				if( k == 1 ) file = file.substring( k );
				else file = file.substring(0, k-1) + file.substring( k );
				k--;					
			}
			else{
				lev++;
				oldk = k;

				if( k >= myParams.getStringCap() &&
						lev >= myParams.getStringCap() ){ 
					//HHHMulti.log("Begrenze vor:  " + file );
					file = file.substring( 0, k );
					//HHHMulti.log("Begrenze nach:  " + file );						
					break;
				}
			}
		}								

		// Schließende Klammern im Dateinamen vermeiden
		String flags = s.substring(pos).split( "[,\\)]" )[1].trim();
		int intFlag  = -1;
		if( flags.contains( "O_RDONLY" ) ) intFlag = 1;
		if( flags.contains( "O_WRONLY" ) ) intFlag = 2;
		if( flags.contains( "O_RDWR"   ) ) intFlag = 3;

		if( ret > -1 ) fdTable.put( new Integer(ret), file );

		//		HHHMulti.log("getOpen: Input " + s + "\nfile= " + file + 
		//				"  flags= " + flags + "  ret= " + ret + 
		//				"  intFlag= " + intFlag );

		if( intFlag < 0 ) 
			throw new RuntimeException("IntFlag nicht gefunden + s");

		String[] eStr = new String[ myParams.getDimS() ];
		eStr[0] = file;		

		String[] eInt = new String[ myParams.getDimI() ];
		for(int i = 1; i < myParams.getDimI(); i++ ) eInt[i] = "";

		eInt[0] = "/1/3/" + intFlag + "/0";
		int[] res = Element.stringToInt( eInt, myParams );
		e = new SysElement( eStr, res, myParams );

		return e;
	}


	// Read und Write sind i.W. identisch
	private static SysElement getReadWrite( String s, Parameter myParams,
			boolean write, HashMap<Integer, String> fdTable ) {

		SysElement e = null;						
		String pay, file = null;		
		String[] par;		
		//int ret = getReturnValue(s);

		// Klammern hinter Gleichheitszeichen ignorieren
		s = s.substring(0, s.lastIndexOf("=") );

		// Klammern im String ignorieren
		int pos = s.indexOf("(");
		int end = s.lastIndexOf(")");
		String parStr = s.substring(pos+1, end);						

		// Komma im String ignorieren
		pos = parStr.indexOf("\"");
		end = parStr.lastIndexOf("\"");		
		if( end > pos && pos > 0 ){
			pay = parStr.substring(pos+1, end);
			parStr = parStr.substring(0, pos) + parStr.substring(end+1);
			par = csv.split( parStr );
		}
		else{
			par = csv.split( parStr );
			pay = par[1];			
		}
		int payHash = 1 + (Math.abs(pay.hashCode()) % 127) ;// Muss > 0 sein, 0 == ANY
		int fd  = Integer.parseInt( par[0].trim() );
		int req = Integer.parseInt( par[2].trim() );
		int size = 1;					// Klein
		if( req > 1024 ) size = 3;		// Mittel
		else if( req > 128  ) size = 2; // Groß

		if( fd > -1 ) file = fdTable.get(fd);
		if( file == null ) file = "/DateiUnbekannt";

		//		HHHMulti.log("getReadWrite: Input " + s + "\nfd= " + fd + 
		//				"  payload= " + pay + "  req= " + req + 
		//				"  ret= " + ret + " size= " +size 
		//				+ " payHash " + payHash + " file= " + file);

		String[] eStr = new String[ myParams.getDimS() ];
		eStr[0] = file;		

		String[] eInt = new String[ myParams.getDimI() ];
		for(int i = 1; i < myParams.getDimI(); i++ ) eInt[i] = "";

		if( write ) eInt[0] = "/1/1/" + size + "/" + payHash;
		else eInt[0] = "/1/2/"  + size + "/" + payHash;

		int[] res = Element.stringToInt( eInt, myParams );
		e = new SysElement( eStr, res, myParams );

		return e;
	}


	private static SysElement getRead(String s, 
			Parameter myParams, HashMap<Integer, String> fdTable) {
		return getReadWrite( s, myParams, false, fdTable );
	}


	private static SysElement getWrite(String s, 
			Parameter myParams, HashMap<Integer, String> fdTable) {
		return getReadWrite( s, myParams, true, fdTable );
	}	


	private static SysElement getSimpleWPath( String s, String code,
			Parameter myParams, HashMap<Integer, String> fdTable ) {

		String[] eStr = new String[ myParams.getDimS() ];
		eStr[0] = getFileName( s, fdTable );
		return getSimpleGeneric( code, eStr, myParams );
	}


	private static SysElement getSimple( String code,
			Parameter myParams) {

		String[] eStr = new String[ myParams.getDimS() ];
		eStr[0] = "*";
		return getSimpleGeneric( code, eStr, myParams );
	}


	private static SysElement getSimpleGeneric( String code, String[] eStr,
			Parameter myParams ) {

		SysElement e = null;
		String[] eInt = new String[ myParams.getDimI() ];
		for(int i = 1; i < myParams.getDimI(); i++ ) eInt[i] = "";
		eInt[0] = code;
		int[] res = Element.stringToInt( eInt, myParams );
		e = new SysElement( eStr, res, myParams );		
		return e;
	}


	private static int getReturnValue(String s){		

		int pos = s.lastIndexOf("=");				
		int end = s.indexOf(" ", pos + 2 );
		if( end <= pos ) end = s.length() ;

		int ret = Integer.parseInt( s.substring(pos+2, end).trim() );
		return ret;
	}


	private static String getFileName(String s, 
			HashMap<Integer, String> fdTable ){		

		String file  = null;
		int pos = s.indexOf("(");		
		int end1 = s.indexOf(",", pos);
		int end2 = s.indexOf(")", pos);
		int end = end1;
		if( end2 > 1 && (end1 > end2 || end1 < 0) ) end = end2;
		int fd  = Integer.parseInt( s.substring(pos+1, end).trim() );		
		if( fd > -1 ) file = fdTable.get(fd);
		if( file == null ) file = "/DateiUnbekannt";
		return file;
	}

}
