package hitters.multi;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;

/**
 * @author Peter Fricke
 * 
 */
public class Taxonomy{

	private static final int maxWidth = 128;		

	private TaxNode root = new TaxNode("*");
	private int size = 0;


	public static void main(String[] args) {

		String label;
		int[] code = new int[SysParser.MAXDEPTH];
		Taxonomy tax = null; 
		try{
			tax  = new Taxonomy("resources/Taxonomy3");
		}catch(java.io.FileNotFoundException e) { 
			try{
				if( tax == null )
					tax  = new Taxonomy("../resources/Taxonomy3");
			}catch(java.io.FileNotFoundException ex) { 
				//System.out.print( ex.getMessage());
			};
		};

		

		for( int i = 0; i < maxWidth; i++ )
			for( int j = 0; j < maxWidth; j++ )
				for( int k = 0; k < maxWidth; k++ ){
					code[1] = i;
					code[2] = j;
					code[3] = k;						
					if( (label = tax.getPathlabel( code )) != null ) {
						System.out.print( "\n" + label + " -> "); 
						for(int m=1; m< 5; m++)
							System.out.print(" /" + code[m]);
					}
				}
	}


	public Taxonomy( String file ) throws java.io.FileNotFoundException{

		String[] components = new String[SysParser.MAXDEPTH];
		String line;
		int[] code = new int[SysParser.MAXDEPTH];
		int level = 0;

		deleteSuffix( code, components, level );

		try {
			BufferedReader in = new BufferedReader( new FileReader(file) );
			while( (line = in.readLine()) != null) {				
				if( line.length() > 2 && ( ! line.startsWith("#") ) ){
					if( getLevel( line ) <= level ){
						add( code, components );
						level = getLevel( line );
						deleteSuffix( code, components, level );
						components[level] = getComponent( line );
						code[level]++;
					}
					else{
						level++;
						components[level] = getComponent( line );
						code[level]++;
					}
				}				
			}
			// Letze Zeile nicht vergessen
			if( level > 0 )	add( code, components );

		} catch( IOException e ) {
			throw new RuntimeException(e);
		}		
	}


	public void add( int[] call, String[] components ) {
		trav( call, components );
		size++;				
	}


	public String getPathlabel( int[] call ) {
		return trav( call, null );
	}


	// Durchlaufe Taxonomie, erzeuge Knoten gdw. components != null
	private String trav( int[] call, String[] components ) {

		if( call == null ) throw new NullPointerException();

		if( components != null && components[1] == "" ) 
			throw new RuntimeException( "Taxonomie: Versuch * einzufuegen");

		if( call[1] == 0 ){
			for( int i = 1; i < call.length; i++ ){
				if( call[i] != 0 ) return null;
			}
			return "*";
		}

		TaxNode node = root;
		String pathlabel = "";
		int index;

		for( int i = 1; i < call.length; i++ ){

			index = call[i];

			if( index < 0 || index >= maxWidth ) {
				throw new RuntimeException( "Taxonomie: " +
						"Zu viele Kinder erwartet. index=" + index +
						"maxWidth= " + maxWidth );
			}
			if( index == 0 ){
				for( int j = i; j < call.length; j++ ){
					if( call[j] != 0 ) return null;
				}
				break;
			}

			if( node.children[index] == null ) {
				if( components != null ) {
					node.children[index] = new TaxNode( components[ i ] );
				}
				// Fuer Hash und andere nicht explizit in der Taxonomie
				// aufgefuehrte Werte
				else{ 
					for( int k = i; k < SysParser.MAXDEPTH; k++ ){
						if( call[ k ] != 0 ) 
							pathlabel = pathlabel + "/" + call[ k ];
					}
					return pathlabel;
				}
			}			
			node = node.children[index];
			pathlabel = pathlabel + "/" + node.label;
		}

		return pathlabel;
	}


	public int size(){
		return this.size;
	}


	private void deleteSuffix( int[] code, String[] components, int level ){
		for( int i = level+1; i < SysParser.MAXDEPTH; i++ ){
			code[i] = 0;
			components[i] = "";
		}
	}

	private String getComponent( String line ){
		int pos = line.lastIndexOf("-");
		return line.substring( pos + 1 ).trim();		
	}

	private int getLevel( String line ){
		int pos = line.lastIndexOf("-");
		for( int i = 0; i < pos; i++ ){
			if( line.charAt(i) != '-' )
				throw new RuntimeException(" Minuszeichen" +
						" fehlen: getLevel:"+ line);
		}		
		return pos + 2;
	}


	private class TaxNode {
		//TODO: Kinder als Liste
		public TaxNode[] children = new TaxNode[maxWidth];
		public String label;
		public TaxNode( String label ){ this.label = label; }
		public String toString(){ return label; }
	}

}